@@ -16,14 +16,16 @@ import pathlib
 from datetime import datetime, timedelta
 from typing import List, Tuple, Literal
 from typing_extensions import TypeAlias
-from typer import Typer
+from typer import Typer, Option
 from circup import find_device
 from tabulate import tabulate
 from circlink.link import LINKS_DIRECTORY, CircuitPythonLink
 
-_TableRowEntry: TypeAlias = Tuple[int, str, bool, pathlib.Path, pathlib.Path, bool, int]
+_TableRowEntry: TypeAlias = Tuple[
+    int, str, bool, pathlib.Path, pathlib.Path, bool, int, str
+]
 
-__version__ = "1.1.2"
+__version__ = "2.1.0"
 
 # Prevent running on non-POSIX systems that don't have os.fork()
 if os.name != "posix":
@@ -51,7 +53,32 @@ def start(
     wipe_dest: bool = False,
     skip_presave: bool = False,
 ) -> None:
-    """Start a CiruitPython link"""
+    """Start a CircuitPython link"""
+
+    _start(
+        read_path,
+        write_path,
+        os.getcwd(),
+        path=path,
+        name=name,
+        recursive=recursive,
+        wipe_dest=wipe_dest,
+        skip_presave=skip_presave,
+    )
+
+
+def _start(
+    read_path: str,
+    write_path: str,
+    base_dir: str,
+    *,
+    path: bool = False,
+    name: str = "",
+    recursive: bool = False,
+    wipe_dest: bool = False,
+    skip_presave: bool = False,
+) -> None:
+    """Backend of starting a CiruitPython link"""
 
     if "*" not in read_path and recursive:
         print("--recursive can only be used with glob patterns!")
@@ -64,9 +91,12 @@ def start(
             sys.exit(1)
         write_path = os.path.join(device_path, write_path)
 
+    base_dir = os.getcwd() if not base_dir else base_dir
+
     link = CircuitPythonLink(
         read_path,
         write_path,
+        base_dir,
         name=name,
         recursive=recursive,
         wipe_dest=wipe_dest,
@@ -108,7 +138,7 @@ def start(
         sys.exit(0)
 
 
-def _stop_link(link_id: int) -> Literal[True]:
+def _stop_link(link_id: int, *, hard_fault: bool = True) -> Literal[True]:
 
     try:
         link = CircuitPythonLink.load_link_by_num(link_id)
@@ -118,7 +148,10 @@ def _stop_link(link_id: int) -> Literal[True]:
 
     if link.stopped:
         print(f"Link #{link.link_id} is already stopped")
-        sys.exit(0)
+        if hard_fault:
+            sys.exit(0)
+        else:
+            return False
 
     link.end_flag = True
     link.save_link()
@@ -131,7 +164,10 @@ def _stop_link(link_id: int) -> Literal[True]:
         time.sleep(0.1)  # Slight delay
         if datetime.now() >= error_time:
             print(f"Link #{link.link_id} could not be stopped!")
-            sys.exit(1)
+            if hard_fault:
+                sys.exit(1)
+            else:
+                return False
 
     print(f"Stopped link #{link_id}")
     return True
@@ -142,13 +178,15 @@ def stop(link_id: str) -> bool:
     """Stop a CircuitPython link"""
 
     if link_id == "all":
-        while stop("last"):
-            pass
-        return True
+        link_entries = _get_links_list("*")
+        for link_entry in link_entries:
+            _stop_link(link_entry[0], hard_fault=False)
+        sys.exit(0)
     if link_id == "last":
         link_id = str(CircuitPythonLink.get_next_link_id() - 1)
         if link_id == "0":
-            return False
+            print("There are no links in the history")
+            sys.exit(1)
 
     try:
         link_id = int(link_id)
@@ -159,7 +197,7 @@ def stop(link_id: str) -> bool:
     return _stop_link(link_id)
 
 
-def _clear_link(link_id: int, *, force: bool = False) -> bool:
+def _clear_link(link_id: int, *, force: bool = False, hard_fault: bool = False) -> bool:
 
     try:
         link = CircuitPythonLink.load_link_by_num(link_id)
@@ -169,8 +207,11 @@ def _clear_link(link_id: int, *, force: bool = False) -> bool:
 
     if not link.stopped and not force:
         print("Can only clear links marked as inactive.")
-        print("To force clear this link, use the --force option.")
-        sys.exit(1)
+        print(f"To force clear link #{link.link_id}, use the --force option.")
+        if hard_fault:
+            sys.exit(1)
+        else:
+            return False
 
     os.remove(link.link_id_to_filename(link_id))
     print(f"Removed link #{link_id} from history")
@@ -183,9 +224,13 @@ def clear(link_id: str, *, force: bool = False) -> None:
     """Clear the link from the history"""
 
     if link_id == "all":
-        while clear("last", force=force):
-            pass
-        return True
+        link_entries = _get_links_list("*")
+        for link_entry in link_entries:
+            _clear_link(link_entry[0], force=force, hard_fault=False)
+        sys.exit(0)
+        # while clear("last", force=force):
+        #    pass
+        # return True
     if link_id == "last":
         link_id = str(CircuitPythonLink.get_next_link_id() - 1)
         if link_id == "0":
@@ -200,27 +245,32 @@ def clear(link_id: str, *, force: bool = False) -> None:
     return _clear_link(link_id, force=force)
 
 
+def _add_links_header() -> List[Tuple[str, ...]]:
+    """Get the header row for the links list"""
+
+    return (
+        "ID",
+        "Name",
+        "Running?",
+        "Read Path",
+        "Write Path",
+        "Recursive?",
+        "Process ID",
+        "Base Directory",
+    )
+
+
 def _get_links_list(
     pattern: str, *, abs_paths: bool = False, name: str = ""
 ) -> List[_TableRowEntry]:
 
     link_paths = pathlib.Path(LINKS_DIRECTORY).glob(pattern)
 
-    link_infos = [
-        (
-            "ID",
-            "Name",
-            "Running?",
-            "Read Path",
-            "Write Path",
-            "Recursive?",
-            "Process ID",
-        )
-    ]
+    link_infos = []
     for link_path in link_paths:
         link = CircuitPythonLink.load_link_by_filepath(str(link_path))
         link_id = link.link_id
-        link_name = "---" if not link.name else link.name
+        link_name = link.name
         link_running = not link.stopped
         if not abs_paths:
             try:
@@ -228,10 +278,11 @@ def _get_links_list(
             except ValueError:
                 abs_paths = True
         if abs_paths:
-            link_read = link.read_path.resolve().absolute()
-        link_write = link.write_path.absolute().resolve()
+            link_read = link.read_path.resolve()
+        link_write = link.write_path.resolve()
         link_recursive = link.recursive
         link_proc = link.process_id
+        link_base = link.base_dir
         if not name or link_name == name:
             link_infos.append(
                 (
@@ -242,19 +293,23 @@ def _get_links_list(
                     link_write,
                     link_recursive,
                     link_proc,
+                    link_base,
                 )
             )
 
-    return link_infos
+    return sorted(link_infos, key=lambda x: x[0])
 
 
-@app.command(name="list")
-def list_links(link_id: str, *, abs_paths: bool = False) -> None:
-    """List all of the tracked links"""
+@app.command()
+def view(link_id: str, *, abs_paths: bool = False) -> None:
+    """List links in the history"""
+
+    last_requested_flag = False
 
     if link_id == "all":
         pattern = "*"
     elif link_id == "last":
+        last_requested_flag = True
         link_id = str(CircuitPythonLink.get_next_link_id() - 1)
         pattern = "link" + link_id + ".json"
         if link_id == "0":
@@ -266,8 +321,18 @@ def list_links(link_id: str, *, abs_paths: bool = False) -> None:
         except ValueError:
             print('Please use a valid link ID, "last", or "all" (default)')
 
-    link_infos = _get_links_list(pattern, abs_paths=abs_paths)
-    print(tabulate(link_infos, headers="firstrow"))
+    link_infos = [x[:-1] for x in _get_links_list(pattern, abs_paths=abs_paths)]
+
+    if not link_infos:
+        if link_id == "all" or last_requested_flag:
+            print("No links in the history to view")
+            sys.exit(0)
+        print("This link ID is not in the history")
+        sys.exit(1)
+
+    # link_infos.insert(0, _add_links_header())
+    show_list = _add_links_header()
+    print(tabulate(link_infos, headers=show_list))
 
 
 @app.command()
@@ -291,18 +356,21 @@ def restart(link_id: str) -> None:
             sys.exit(1)
 
     link_list = _get_links_list(pattern)
-    if len(link_list) == 1:
+    if not link_list:
         print("There are no links in the history to restart")
         sys.exit(1)
 
-    for index, link in enumerate(link_list):
-        if not index:
-            continue
+    for link in link_list:
         if link[2]:
             print(f"Link #{link[0]} is active, not restarting this link.")
         else:
-            start(
-                str(link[3]), str(link[4]), name=link[1], recursive=link[5], path=True
+            _start(
+                str(link[3]),
+                str(link[4]),
+                link[-1],
+                name=link[1],
+                recursive=link[5],
+                path=True,
             )
             clear(link[0])
 
@@ -316,7 +384,6 @@ def detect() -> None:
         print("CircuitPython device detected:", device)
     else:
         print("No CircuitPython device detected")
-    sys.exit(0)
 
 
 @app.command()
@@ -325,7 +392,6 @@ def about() -> None:
 
     print("Originally built with love by Tekktrik")
     print("Happy hackin'!")
-    sys.exit(0)
 
 
 @app.command()
@@ -333,7 +399,6 @@ def version() -> None:
     """Display the current version of circlink"""
 
     print(__version__)
-    sys.exit(0)
 
 
 def main() -> None:
